from si_tests.utils import packaging_version as version
from si_tests.utils import utils

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from si_tests.managers.kaas_manager import Cluster


class WorkaroundCheckManagerBase(object):
    """Common methods useful for workarounds check"""

    def __init__(self, cluster: "Cluster"):
        self._cluster: "Cluster" = cluster

    @property
    def cluster(self) -> "Cluster":
        return self._cluster

    @property
    def k8sclient(self):
        return self._cluster.k8sclient

    @property
    def mcc_version(self):
        """Active MCC release version"""
        mgmt_cluster = self._cluster._manager.get_mgmt_cluster()
        return mgmt_cluster.get_kaasrelease_version()

    @property
    def clusterrelease_version(self):
        """Current Cluster version"""
        return self._cluster.clusterrelease_version


class WorkaroundCheckManager(WorkaroundCheckManagerBase):
    """Workaround check manager

    Purpose of this class is to keep all the checks for workarounds in a single place.

    - Each method should return 'True' if the related workaround should be applied.
      Please don't add any waiters/actions here. Only checks for versions/providers/cluster objects/etc...

    - Methods should be named using PRODX bugs/tasks, for example:

          def prodx_12345(self):

    - Please clear the code from outdated workarounds after the related tasks are resolved
    """

    def prodx_34281(self):
        # https://mirantis.jira.com/browse/PRODX-34281 , fix version: 2.24.3
        if version.parse(self.mcc_version) < version.parse("kaas-2-24-3") and self.cluster.provider in [
                utils.Provider.baremetal, utils.Provider.equinixmetalv2, utils.Provider.vsphere]:
            # apply workaround for PRODX-34281 if MCC release is lower than 2.24.3
            return True

        return False

    def field_6283(self):
        # resolve https://mirantis.jira.com/browse/FIELD-6283 planned for 3.6.7 MKE release
        if version.parse(self.clusterrelease_version) < version.parse('mke-16-0-0-3-6-7'):
            return True
        else:
            return False

    def field_5850(self):
        # https://mirantis.jira.com/browse/FIELD-5850
        # current status is pending investigation
        return True

    def prodx_35225(self):
        # https://mirantis.jira.com/browse/PRODX-35225
        # Due to feature is available only starting from mosk-15-0-3-23-2-2
        # we need to rely on this in update scenarios, because mosk-15-0-1 will
        # be supported for a long time without this feature.
        if version.parse(self.clusterrelease_version) < version.parse('mosk-15-0-3-23-2-2'):
            return True
        return False

    def prodx_26477(self):
        # https://mirantis.jira.com/browse/PRODX-26477
        # In-tree vSphere CCM is deprecated and doesn’t work properly with vSphere 6x since k8s 1.25
        # In 2.25 release:
        # vsphere child can be deployed\updated with releases:
        # mke-14-0-1-3-6-5, mke-14-0-4-3-6-6, mke-14-1-0-3-6-6
        # openstack child can be deployed\updated with releases:
        # mke-14-0-1-3-6-5, mke-14-0-4-3-6-6, mke-16-0-0-3-7-1
        # be supported for a long time without this feature.
        if version.parse("kaas-2-25-0-rc") <= version.parse(self.mcc_version) <= version.parse("kaas-2-26-0"):
            return True
        return False

    def prodx_37301(self):
        # https://mirantis.jira.com/browse/PRODX-37301
        # We need to completely exclude mke-14-X versions from deploy
        # starting from 2.25.1
        if version.parse(self.mcc_version) > version.parse("kaas-2-25-0"):
            return True
        return False

    def prodx_27814(self):
        # workaround regarding https://mirantis.jira.com/browse/PRODX-27814
        if version.parse(self.mcc_version) >= version.parse("kaas-2-25-0-rc"):
            if (self.clusterrelease_version.startswith("mke-14")
                    and version.parse(self.clusterrelease_version) >= version.parse('mke-14-1-0-3-6-6')):
                return True
            elif (self.clusterrelease_version.startswith("mke-16")
                  and version.parse(self.clusterrelease_version) >= version.parse('mke-16-0-0-3-7-1')):
                return True
            else:
                return False
        else:
            return False

    def prodx_36209(self):
        # https://mirantis.jira.com/browse/PRODX-36209
        if version.parse(self.clusterrelease_version) >= version.parse('mosk-17-0-0-23-3'):
            return True
        return False

    def prodx_37306(self):
        # https://mirantis.jira.com/browse/PRODX-37306
        if (('mke-16' in self.clusterrelease_version and
             version.parse(self.clusterrelease_version) >= version.parse('mke-16-0-1-3-7-2')) or
                ('mosk' in self.clusterrelease_version and
                 version.parse(self.clusterrelease_version) >= version.parse('mosk-17-0-1-23-3-1'))):
            return True
        return False
